@@ -4,7 +4,7 @@ TBD
 """
 import threading
 import time
-from entangle.logging.debug import logging
+from entangle.logging.info import logging
 from entangle.dataflow import thread
 from entangle.dataflow import process
 from entangle.dataflow import dataflow
